"""The HEPi package aims to automize cluster computations for parameter scans with the option to produce plots."""
from importlib.metadata import version

from .input import *
from .interpolate import *
from .load import *
from .output import *

# TODO remvoe this for speed
from .plot import *
from .results import *
from .run import *
from .util import *

# import pkg_resources as pkg  # part of setuptools
# import json
# import requests
# from urllib.request import urlopen
from ._version import version as __version__

package = "hepi"
__all__ = ["__version__"]


# try:
#    version = pkg.require(package)[0].version
# except pkg.DistributionNotFound:
#    version = "dirty"
# __version__ = version
#



from hepi.results import Result, add_errors, asym_to_sym_error


class NNLLFastResult(Result):
    """
    NNLL-fast Result Data.
    """

    def __init__(
        self,
        NLO,
        NLO_PDF_ERRMINUS,
        NLO_PDF_ERRPLUS,
        aNNLO_PLUS_NNLL,
        aNNLO_PLUS_NNLL_SCALE_ERRMINUS,
        aNNLO_PLUS_NNLL_SCALE_ERRPLUS,
        aNNLO_PLUS_NNLL_PDF_ERRMINUS,
        aNNLO_PLUS_NNLL_PDF_ERRPLUS,
    ):
        Result.__init__(self, None, NLO, None,aNNLO_PLUS_NNLL)
        self.NLO_NOERR= NLO
        self.NLO_PDF_ERRMINUS  =NLO_PDF_ERRMINUS
        self.NLO_PDF_ERRPLUS   =NLO_PDF_ERRPLUS
        self.NLO_PDF = asym_to_sym_error(NLO, NLO_PDF_ERRMINUS, NLO_PDF_ERRPLUS)

        self.aNNLO_PLUS_NNLL_NOERR = aNNLO_PLUS_NNLL

        self.aNNLO_PLUS_NNLL_SCALE_ERRMINUS=aNNLO_PLUS_NNLL_SCALE_ERRMINUS
        self.aNNLO_PLUS_NNLL_SCALE_ERRPLUS =aNNLO_PLUS_NNLL_SCALE_ERRPLUS
        self.aNNLO_PLUS_NNLL_SCALE        = asym_to_sym_error(aNNLO_PLUS_NNLL, aNNLO_PLUS_NNLL_SCALE_ERRMINUS, aNNLO_PLUS_NNLL_SCALE_ERRPLUS)

        self.aNNLO_PLUS_NNLL_PDF_ERRMINUS  =aNNLO_PLUS_NNLL_PDF_ERRMINUS
        self.aNNLO_PLUS_NNLL_PDF_ERRPLUS   =aNNLO_PLUS_NNLL_PDF_ERRPLUS
        self.aNNLO_PLUS_NNLL_PDF           = asym_to_sym_error(aNNLO_PLUS_NNLL, aNNLO_PLUS_NNLL_PDF_ERRMINUS, aNNLO_PLUS_NNLL_PDF_ERRPLUS)

        self.aNNLO_PLUS_NNLL_ERRMINUS = -add_errors(aNNLO_PLUS_NNLL_SCALE_ERRMINUS, aNNLO_PLUS_NNLL_PDF_ERRMINUS)
        self.aNNLO_PLUS_NNLL_ERRPLUS= add_errors(aNNLO_PLUS_NNLL_SCALE_ERRPLUS, aNNLO_PLUS_NNLL_PDF_ERRPLUS)
        self.aNNLO_PLUS_NNLL = asym_to_sym_error(aNNLO_PLUS_NNLL, self.aNNLO_PLUS_NNLL_ERRMINUS, self.aNNLO_PLUS_NNLL_ERRPLUS) 





import os
import subprocess
import warnings
from typing import List

from hepi.input import Input, update_slha
from hepi.run import Runner


class FeynHiggsRunner(Runner):
    def _check_path(self) -> bool:
        if os.path.exists(os.path.expanduser(self.get_path() + "/bin/FeynHiggs")):
            self.set_path(self.get_path() + "/bin/FeynHiggs")
            return True
        if self.get_path().endswith("FeynHiggs"):
            return True
        return False

    def run(self, slhas: List[Input], **kwargs) -> List[Input]:
        """
        Run the passed list of parameters for feynhiggs.

        Args:
            slhas (:obj:`list` of :class:`Input`): Input parameters with a SLHA file that can be processed by feynhiggs.
        Returns:
            :obj:`list` of :class:`Input`
        """
        if not self._check_path():
            warnings.warn("The path is not valid for " + self.get_name())
        for s in slhas:
            # Remove Creation time for hash-/caching
            comm = (
                "cp "
                + self.get_output_dir()
                + s.slha
                + " feynhiggs_tmp.in && "
                + get_path()
                + " feynhiggs_tmp.in "
                + " && mv "
                + "feynhiggs_tmp.in.fh-001 "
                + self.get_output_dir()
                + s.slha
#                + " && sed -i '/Created/d' "
#                + self.get_output_dir()
#                + s.slha
            )
            #print(comm)
            proc = subprocess.Popen(
                comm, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
            proc.wait()
            # get proc return code
            if proc.returncode != 0:
                warnings.warn("feynhiggs failed")
            update_slha(s)
        return slhas


# Backward compatibility
spheno_default_runner = FeynHiggsRunner("FeynHiggs")
"""Default SoftSusy Runner to provide backward compatibility"""
run = spheno_default_runner.run
set_path = spheno_default_runner.set_path
get_path = spheno_default_runner.get_path



import os
import pkgutil
import stat
import warnings
from string import Template
from typing import List

import pyslha
from uncertainties import ufloat

from hepi.input import Input, Order, is_gluino, is_squark
from hepi.results import Result
from hepi.run import Runner, RunParam
from hepi.run.nllfast.result import NLLFastResult


class NLLfastRunner(Runner):
    def get_version(self) -> str:
        return "3.1"

    def orders(self) -> List[Order]:
        return [Order.LO, Order.NLO, Order.NLO_PLUS_NLL]

    def _get_nf_proc(self, p: Input):
        d = pyslha.read(self.get_output_dir() + p.slha)
        mg = d.blocks["MASS"][1000021]
        ms = 0.0
        for r in range(1000001, 1000006):
            ms += d.blocks["MASS"][r]
        for r in range(2000001, 2000006):
            ms += d.blocks["MASS"][r]
        ms /= 10  # 10 flavors no (s)top
        if p.has_gluino() and p.has_squark():
            if is_squark(p.particle1):
                ms = d.blocks["MASS"][abs(p.particle1)]
            if is_squark(p.particle2):
                ms = d.blocks["MASS"][abs(p.particle2)]
            return "sg", "cteq", ms, mg, 10
        if is_gluino(p.particle1) and is_gluino(p.particle2):
            if ms > 2000:  # go into decoupling limit
                return "gdcpl", "cteq", "", mg, 1
            return "gg", "cteq", ms, mg, 1
        if is_squark(p.particle1) and is_squark(p.particle2):
            ms = (
                d.blocks["MASS"][abs(p.particle1)] + d.blocks["MASS"][abs(p.particle2)]
            ) / 2
            if mg > 2000:  # go into decoupling limit
                return "sdcpl", "cteq", ms, "", 10
            if p.particle1 > 0 and p.particle2 > 0:
                return "ss", "cteq", ms, mg, 10 * 10
            elif (p.particle1 > 0 and p.particle2 < 0) or (
                p.particle1 < 0 and p.particle2 > 0
            ):
                s = p.particle1 if p.particle1 > p.particle2 else p.particle2
                b = p.particle2 if p.particle1 > p.particle2 else p.particle1
                return "sb", "cteq", ms, mg, 10
        return "UNKNOWN_PROCESS_OR_UNIMPLEMENTED_PROCESS"

    def _get_nf_input(self, p: Input) -> dict:
        d = {}
        # d["ps_inlo"] = int(p.order)
        (
            d["nf_final_state_in"],
            d["nf_pdf"],
            d["nf_squark_mass"],
            d["nf_gluino_mass"],
            d["nf_deg"],
        ) = self._get_nf_proc(p)
        return d

    def _check_input(self, p: Input, **kwargs) -> bool:
        """Checks input parameter for compatibility with Prospino"""
        if p.energy != 13000:
            warnings.warn("NLL-fast does not support other energies than 13 TeV.")
            return False
        if p.mu_f != 1.0 or p.mu_r != 1.0:
            warnings.warn("NLL-fast does not support varying the scales manually.")
            return False
        if p.pdf_lo != "cteq6l1" or p.pdf_nlo != "cteq66":
            warnings.warn(
                "NLL-fast does not support all pdfs (CTEQ6L1 and CTEQ66 allowed defaults)."
            )
            return False
        return True

    def _is_valid(self, file: str, p: Input, d, **kwargs) -> bool:
        return super()._is_valid(file, p, d)

    def _parse_file(self, file: str) -> Result:
        # TODO parse result
        ret = []
        with open(file) as output:
            for line in output:
                pass
            for s in line.split():
                if "TOO" in s:
                    warnings.warn(
                        "NLL-fast failed to calculate the cross section due to too large masses."
                    )
                ret.append(float(s))
        return NLLFastResult(  # divide by 10 due to degeneracy, this is injeted into the result
            ret[len(ret) - 14 + 2] / ret[len(ret) - 14 + 13],
            ret[len(ret) - 14 + 3] / ret[len(ret) - 14 + 13],
            -((ret[len(ret) - 14 + 8] ** 2 + ret[len(ret) - 14 + 10] ** 2) ** 0.5 / 100)
            * ret[len(ret) - 14 + 3]
            / ret[len(ret) - 14 + 13],
            ((ret[len(ret) - 14 + 7] ** 2 + ret[len(ret) - 14 + 9] ** 2) ** 0.5 / 100)
            * ret[len(ret) - 14 + 3]
            / ret[len(ret) - 14 + 13],
            ret[len(ret) - 14 + 4] / ret[len(ret) - 14 + 13],
            ret[len(ret) - 14 + 6] / ret[len(ret) - 14 + 13],
            ret[len(ret) - 14 + 5] / ret[len(ret) - 14 + 13],
            -((ret[len(ret) - 14 + 8] ** 2 + ret[len(ret) - 14 + 10] ** 2) ** 0.5 / 100)
            * ret[len(ret) - 14 + 4]
            / ret[len(ret) - 14 + 13],
            ((ret[len(ret) - 14 + 7] ** 2 + ret[len(ret) - 14 + 9] ** 2) ** 0.5 / 100)
            * ret[len(ret) - 14 + 4]
            / ret[len(ret) - 14 + 13],
        )

    def _prepare(self, p: Input, **kwargs) -> RunParam:
        rp = super()._prepare(p, **kwargs)
        if not rp.skip:
            d = p.__dict__

            for k, v in self._get_nf_input(p).items():
                d[k] = v

            with open(rp.execute, "w") as tmp:
                tmp.write(
                    "#!/bin/sh\n"
                    + 'pushd {path} > /dev/null\nR="$({exec} {proc} {pdf} {sq} {gl})"\npopd > /dev/null\necho "$R {deg}">{out}'.format(
                        exec=self.get_path(),
                        path=os.path.dirname(self.get_path())
                        if not os.path.dirname(self.get_path()) == ""
                        else ".",
                        out=rp.out_file,
                        proc=d["nf_final_state_in"],
                        sq=d["nf_squark_mass"],
                        gl=d["nf_gluino_mass"],
                        pdf=d["nf_pdf"],
                        deg=d["nf_deg"],
                    )
                )
            st = os.stat(rp.execute)
            os.chmod(rp.execute, st.st_mode | stat.S_IEXEC)

            sname = d["slha"]
            with open(self.get_output_dir() + sname, "r") as f:
                with open(rp.out_file, "a") as a:
                    a.write(f.read() + "\n\n")
        return rp


# Legacy
default_nllfast_runner = NLLfastRunner("~/git/nll-fast/nll-fast")
"""Default NLL-fast Runner to provide backward compatibility"""
run = default_nllfast_runner.run
set_path = default_nllfast_runner.set_path
get_path = default_nllfast_runner.get_path



import copy
import json
from re import I

from uncertainties import ufloat

from hepi.input import Input, order_to_string, xsec_to_order
from hepi.util import DL2DF, LD2DL


def load_json_with_metadata(file):

    """
    Load xsec data from json in to something that works within hepi's plotting.

    Args:
        f : readable object, e.g. `open(filepath:str)`.
        dimensions (int) : 1 or 2 currently supported.
    """
    dict = json.load(file)

    inpu = Input(
        xsec_to_order(dict["order"]),
        float(dict["Ecom [GeV]"]),
        -1,
        -1,
        slha=dict["process_latex"],
        pdf_lo=dict["PDF set"],
        pdf_nlo=dict["PDF set"],
        update=False,
    )
    so = order_to_string(inpu.order)
    inpu.runner = dict["tool"]
    dat = []
    params = dict["parameters"]
    dimensions = len(params)
    if dimensions == 2:
        for k in dict["data"]:
            for l in dict["data"][k]:
                dicd = copy.copy(inpu.__dict__)
                for p1 in params[0]:
                    dicd[p1] = float(k)
                for p2 in params[1]:
                    dicd[p2] = float(l)
                if "unc_pb" in dict["data"][k][l]:
                    dicd[so + "_NOERR"] = dict["data"][k][l]["xsec_pb"]
                    dicd[so + "_COMBINED"] = ufloat(
                        dict["data"][k][l]["xsec_pb"], dict["data"][k][l]["unc_pb"]
                    )
                elif (
                    "unc_down_pb" in dict["data"][k][l]
                    and "unc_up_pb" in dict["data"][k][l]
                ):
                    dicd[so + "_NOERR"] = dict["data"][k][l]["xsec_pb"]
                    dicd[so + "_COMBINED"] = ufloat(
                        dict["data"][k][l]["xsec_pb"]
                        + (
                            dict["data"][k][l]["unc_up_pb"]
                            + dict["data"][k][l]["unc_down_pb"]
                        )
                        / 2,
                        (
                            dict["data"][k][l]["unc_up_pb"]
                            - dict["data"][k][l]["unc_down_pb"]
                        )
                        / 2,
                    )
                    if (
                        "unc_scale_up_pb" in dict["data"][k][l]
                        and "unc_scale_down_pb" in dict["data"][k][l]
                    ):
                        dicd[so + "_SCALE"] = ufloat(
                            dict["data"][k][l]["xsec_pb"]
                            + (
                                dict["data"][k][l]["unc_scale_up_pb"]
                                + dict["data"][k][l]["unc_scale_down_pb"]
                            )
                            / 2,
                            (
                                dict["data"][k][l]["unc_scale_up_pb"]
                                - dict["data"][k][l]["unc_scale_down_pb"]
                            )
                            / 2,
                        )
                    if (
                        "unc_pdf_up_pb" in dict["data"][k][l]
                        and "unc_pdf_down_pb" in dict["data"][k][l]
                    ):
                        dicd[so + "_PDF"] = ufloat(
                            dict["data"][k][l]["xsec_pb"]
                            + (
                                dict["data"][k][l]["unc_pdf_up_pb"]
                                + dict["data"][k][l]["unc_pdf_down_pb"]
                            )
                            / 2,
                            (
                                dict["data"][k][l]["unc_pdf_up_pb"]
                                - dict["data"][k][l]["unc_pdf_down_pb"]
                            )
                            / 2,
                        )
                else:
                    raise ValueError("No uncertainty found in data.")
                dat.append(dicd)
    if dimensions == 1:
        for k in dict["data"]:
            dicd = copy.copy(inpu.__dict__)
            for p1 in params[0]:
                dicd[p1] = float(k)
            if "unc_pb" in dict["data"][k]:
                dicd[so + "_NOERR"] = dict["data"][k]["xsec_pb"]
                dicd[so + "_COMBINED"] = ufloat(
                    dict["data"][k]["xsec_pb"], dict["data"][k]["unc_pb"]
                )
            elif "unc_down_pb" in dict["data"][k] and "unc_up_pb" in dict["data"][k]:
                dicd[so + "_NOERR"] = dict["data"][k]["xsec_pb"]
                dicd[so + "_COMBINED"] = ufloat(
                    dict["data"][k]["xsec_pb"]
                    + (dict["data"][k]["unc_up_pb"] + dict["data"][k]["unc_down_pb"])
                    / 2,
                    (dict["data"][k]["unc_up_pb"] - dict["data"][k]["unc_down_pb"]) / 2,
                )
                if (
                    "unc_scale_up_pb" in dict["data"][k]
                    and "unc_scale_down_pb" in dict["data"][k]
                ):
                    dicd[so + "_SCALE"] = ufloat(
                        dict["data"][k]["xsec_pb"]
                        + (
                            dict["data"][k]["unc_scale_up_pb"]
                            + dict["data"][k]["unc_scale_down_pb"]
                        )
                        / 2,
                        (
                            dict["data"][k]["unc_scale_up_pb"]
                            - dict["data"][k]["unc_scale_down_pb"]
                        )
                        / 2,
                    )
                if (
                    "unc_pdf_up_pb" in dict["data"][k]
                    and "unc_pdf_down_pb" in dict["data"][k]
                ):
                    dicd[so + "_PDF"] = ufloat(
                        dict["data"][k]["xsec_pb"]
                        + (
                            dict["data"][k]["unc_pdf_up_pb"]
                            + dict["data"][k]["unc_pdf_down_pb"]
                        )
                        / 2,
                        (
                            dict["data"][k]["unc_pdf_up_pb"]
                            - dict["data"][k]["unc_pdf_down_pb"]
                        )
                        / 2,
                    )
            else:
                raise ValueError("No uncertainty found in data.")
            dat.append(dicd)
    ddf = DL2DF(LD2DL(dat, actual_dict=True))
    # copy mass list to single masses
    for k in ddf.keys():
        if k.startswith("mass_"):
            for m in k[5:].split("_"):
                ddf["mass_" + m] = ddf[k]
    return ddf, params, inpu


def load_json(f, dimensions=1):
    return load_json_with_metadata(f)[0]


load = load_json



from typing import List, Tuple

from particle import PDGID
from particle.converters.bimap import DirectionalMaps

PDG2LaTeXNameMap, LaTeX2PDGNameMap = DirectionalMaps(
    "PDGID", "LaTexName", converters=(PDGID, str)
)

PDG2Name2IDMap, PDGID2NameMap = DirectionalMaps(
    "PDGName", "PDGID", converters=(str, PDGID)
)


def get_name(pid: int) -> str:
    """
    Get the latex name of a particle.

    Args:
        pid (int) : PDG Monte Carlo identifier for the particle.

    Returns:
        str: Latex name.

    Examples:
        >>> get_name(21)
        'g'
        >>> get_name(1000022)
        '\\\\tilde{\\\\chi}_{1}^{0}'
    """
    global PDG2LaTeXNameMap
    pdgid = PDG2LaTeXNameMap[pid]
    return pdgid


def get_LR_partner(pid: int) -> Tuple[int, int]:
    """Transforms a PDG id to it's left-right partner.

    Args:
        pid (int) : PDG Monte Carlo identifier for the particle.

    Returns:
        tuple : First int is -1 for Left and 1 for Right. Second int is the PDG id.

    Examples:
        >>> get_LR_partner(1000002)
        (-1, 2000002)
    """
    global PDG2Name2IDMap
    n = PDGID2NameMap[pid]
    if "L" in n:
        n = n.replace("L", "R")
        return -1, int(PDG2Name2IDMap[n])
    if "R" in n:
        n = n.replace("R", "L")
        return 1, int(PDG2Name2IDMap[n])
    if "1" in n:
        n = n.replace("1", "2")
        return -1, int(PDG2Name2IDMap[n])
    if "2" in n:
        n = n.replace("2", "1")
        return 1, int(PDG2Name2IDMap[n])
    return None



"""
Collection of utility functions for the :mod:`hepi` package.
"""
import hashlib
import json
import warnings
from typing import List, Tuple

import numpy as np
import pandas as pd


class DictData:
    def __str__(self):
        """Returns attributes as dict as string"""
        return str(self.__dict__)


def LD2DL(l: List, actual_dict=False) -> dict:
    """
    Convert a list of objects into a dictionary of lists.

    The values of each object are first converted to a `dict` through the `__dict__` attribute.

    Args:
        l (List) : list of objects.
        actual_dict (bool) : objects are already dicts

    Returns:
        dict : dictionary of numpy arrays.

    Examples:
        >>> class Param:
        ...      def __init__(self,a,b,c):
        ...         self.a = a
        ...         self.b = b
        ...         self.c = c
        >>> LD2DL([ Param(1,2,3), Param(4,5,6) , Param(7,8,9) ])
        {'a': array([1, 4, 7]), 'b': array([2, 5, 8]), 'c': array([3, 6, 9])}
    """
    # Check l[0] keys in all dictionaries.
    for m in l:
        md = m if actual_dict else m.__dict__
        for k in l[0] if actual_dict else l[0].__dict__:
            assert k in md
    # switch them
    return {
        k: np.array([dic[k] if actual_dict else dic.__dict__[k] for dic in l])
        for k in (l[0] if actual_dict else l[0].__dict__)
    }


def DL2DF(ld: dict) -> pd.DataFrame:
    """
    Convert a `dict` of `list`s to a `pandas.DataFrame`.
    """
    return pd.DataFrame.from_dict(ld)


def namehash(n: any) -> str:
    """
    Creates a sha256 hash from the objects string representation.

    Args:
        n (any) : object.

    Returns:
        str: sha256 of object.

    Examples:
        >>> p = {'a':1,'b':2}
        >>> str(p)
        "{'a': 1, 'b': 2}"
        >>> namehash(str(p))
        '3dffaea891e5dbadb390da33bad65f509dd667779330e2720df8165a253462b8'
        >>> namehash(p)
        '3dffaea891e5dbadb390da33bad65f509dd667779330e2720df8165a253462b8'
    """
    m = hashlib.sha256()
    m.update(str(n).encode("utf-8"))
    return m.hexdigest()


def lhapdf_name_to_id(name: str) -> int:
    """
    Converts a LHAPDF name to the sets id.

    Args:
        name (str) : LHAPDF set name.

    Returns:
        int: id of the LHAPDF set.

    Examples:
        >>> lhapdf_name_to_id("CT14lo")
        13200
    """
    try:
        import lhapdf
    except ImportError:
        warnings.warn(
            "LHAPDF python binding not installed? Make sure you set PYTHONPATH correctly (i.e. correct python version) if you want to compute PDF uncertainties."
        )
        return 0
    if not lhapdf.availablePDFSets():
        warnings.warn(
            "No PDF sets found. Make sure the environment variable LHAPDF_DATA_DIR points to the correct location (.../share/LHAPDF)."
        )
        return 0
    if not name in lhapdf.availablePDFSets():
        warnings.warn("PDF set '" + name + "' not installed?")
        return 0
    return lhapdf.getPDFSet(name).lhapdfID


def lhapdf_id_to_name(lid: int) -> str:
    try:
        import lhapdf
    except ImportError:
        warnings.warn(
            "LHAPDF python binding not installed? Make sure you set PYTHONPATH correctly (i.e. correct python version) if you want to compute PDF uncertainties.."
        )
        return ""
    if not lhapdf.availablePDFSets():
        warnings.warn(
            "No PDF sets found. Make sure the environment variable LHAPDF_DATA_DIR points to the correct location (.../share/LHAPDF)."
        )
        return 0
    for n in lhapdf.availablePDFSets():
        if lhapdf.getPDFSet(n).lhapdfID == lid:
            return n

    warnings.warn("PDF set with id " + str(lid) + " unknown/not installed?")
    return "Unknown PDF ID: " + str(lid)



"""Runs Resummino"""
import os
import os.path
import pkgutil
import stat
import warnings
from string import Template
from typing import List

from hepi.input import Order
from hepi.results import Result
from hepi.run import Runner, RunParam

from .. import Input
from .result import is_valid, parse_single


class ResumminoRunner(Runner):
    def orders(self) -> List[Order]:
        return [Order.LO, Order.NLO, Order.NLO_PLUS_NLL, Order.aNNLO_PLUS_NNLL]

    def get_version(self) -> str:
        p = os.path.expanduser(self.get_path())
        ret = self._sub_run([p, "--version"])
        return ret.split("\n")[-2]

    def _check_path(self) -> bool:
        if os.path.exists(os.path.expanduser(self.get_path() + "/build/bin/resummino")):
            self.set_path(self.get_path() + "/build/bin/resummino")
            return True
        if os.path.exists(os.path.expanduser(self.get_path() + "/bin/resummino")):
            self.set_path(self.get_path() + "/bin/resummino")
            return True
        if self.get_path().endswith("resummino"):
            return True
        return False

    def _check_input(self, p: Input, **kwargs) -> bool:
        if p.order == Order.aNNLO_PLUS_NNLL and (
            (p.has_gluino() and p.has_weakino()) or (p.has_squark() and p.has_weakino())
        ):
            warnings.warn("Resummino does not support stong-weak mixed aNNLO+NNLL.")
            return False
        return True

    def _is_valid(self, file: str, p: Input, d, **kwargs) -> bool:
        if not super()._is_valid(file, p, d):
            return False
        return is_valid(file, p, d, **kwargs)

    def _parse_file(self, file: str) -> Result:
        return parse_single(file)

    def _prepare(self, p: Input, **kwargs) -> RunParam:
        rp = super()._prepare(p, **kwargs)
        if not rp.skip:
            data = pkgutil.get_data(__name__, "plot_template.in").decode("utf-8")
            flags = ""
            if p.order == Order.LO:
                flags = flags + "--lo"
            elif p.order == Order.NLO:
                flags = flags + "--nlo"
            elif p.order == Order.NLO_PLUS_NLL:
                flags = flags + "--nll"
            elif p.order == Order.aNNLO_PLUS_NNLL:
                flags = flags + "--nnll"
            else:
                raise ValueError(
                    "Order not supported by resummino. Must be one of LO/NLO/NLO+NLL/aNNLO+NNLL."
                )

            src = Template(data)
            result = src.substitute(p.__dict__)
            od = self.get_output_dir()
            if not os.path.exists(od):
                os.makedirs(od)
            with open(od + rp.name + ".in", "w") as tmp:
                tmp.write(result)
            with open(od + rp.name + ".sh", "w") as tmp:
                tmp.write(
                    "#!/bin/sh\n"
                    + get_path()
                    + " {} {} >> {}".format(
                        od + rp.name + ".in", flags, od + rp.name + ".out"
                    )
                )
            st = os.stat(od + rp.name + ".sh")
            os.chmod(od + rp.name + ".sh", st.st_mode | stat.S_IEXEC)
            with open(od + rp.name + ".out", "w") as tmp:
                tmp.write(result + "\n\n")

            sname = p.slha
            with open(od + sname, "r") as f:
                with open(od + rp.name + ".out", "a") as a:
                    a.write(f.read() + "\n\n")
        return rp


# Legacy
default_resummino_runner = ResumminoRunner("~/resummino/")
"""Default Resummino Runner to provide backward compatibility"""
run = default_resummino_runner.run
set_path = default_resummino_runner.set_path
get_path = default_resummino_runner.get_path
get_version = default_resummino_runner.get_version



"""Results and postprocessing for the :mod:`hepi` package."""
import multiprocessing as mp
import warnings

import numpy as np
import uncertainties.unumpy as unp
from uncertainties import ufloat

# from pqdm.processes import pqdm as ppqdm
from pqdm.threads import pqdm as tpqdm

from .util import DictData

# If the numerical uncertainty times :attr:`required_numerical_uncertainty_factor` is higher than the scale or pdf uncertainty a warning is shown.
required_numerical_uncertainty_factor = 5

unv = unp.nominal_values
usd = unp.std_devs


class Result(DictData):
    """
    General result class. All uncertainties are of numerical origin.

    Attributes:
        LO (:obj:`double`): Leading Order result. Defaults to None.
        NLO (:obj:`double`): Next-to-Leading Order result. Defaults to None.
        NLO_PLUS_NLL (:obj:`double`): Next-to-Leading Order plus Next-to-Leading Logarithm result. Defaults to None.
        K_LO (:obj:`double`): LO divided by LO.
        K_NLO (:obj:`double`): NLO divided by LO result.
        K_NLO_PLUS_NLL (:obj:`double`): NLO+NLL divided by LO.
        K_aNNLO_PLUS_NNLL (:obj:`double`): aNNLO+NNLL divided by LO.
        NLO_PLUS_NLL_OVER_NLO (:obj:`double`): NLO+NLL divided by NLO.
        aNNLO_PLUS_NNLL_OVER_NLO (:obj:`double`): aNNLO+NNLL divided by NLO.
    """

    def __init__(self, lo=None, nlo=None, nlo_plus_nll=None, annlo_plus_nnll=None):
        """
        Sets given and computes dependent ``Attributes``.

        Args:
            lo (:obj:`double`): corresponds to :attr:`LO`.
            nlo (:obj:`double`): corresponds to :attr:`NLO`.
            nlo_plus_nll (:obj:`double`): corresponds to :attr:`NLO_PLUS_NLL`.
            annlo_plus_nnll (:obj:`double`): corresponds to :attr:`aNNLO_PLUS_NNLL`.
        """
        self.LO = lo
        self.NLO = nlo
        self.NLO_PLUS_NLL = nlo_plus_nll
        self.aNNLO_PLUS_NNLL = annlo_plus_nnll
        if lo is not None and lo != 0:
            self.K_LO = lo / lo
        else:
            self.K_LO = None
        if nlo is not None and lo != 0 and lo is not None:
            self.K_NLO = nlo / lo
        else:
            self.K_NLO = None
        if nlo_plus_nll is not None and lo != 0 and lo is not None:
            self.K_NLO_PLUS_NLL = nlo_plus_nll / lo
        else:
            self.K_NLO_PLUS_NLL = None

        if nlo_plus_nll is not None and nlo != 0 and nlo is not None:
            self.NLO_PLUS_NLL_OVER_NLO = nlo_plus_nll / nlo
        else:
            self.NLO_PLUS_NLL_OVER_NLO = None

        if annlo_plus_nnll is not None and lo != 0 and lo is not None:
            self.K_aNNLO_PLUS_NNLL = annlo_plus_nnll / lo
        else:
            self.K_aNNLO_PLUS_NNLL = None

        if annlo_plus_nnll is not None and nlo != 0 and nlo is not None:
            self.aNNLO_PLUS_NNLL_OVER_NLO = annlo_plus_nnll / nlo
        else:
            self.aNNLO_PLUS_NNLL_OVER_NLO = None


# TODO detect which errors also for scales
def pdf_errors(li, dl, ordernames=None, confidence_level=90, n_jobs=None):
    """
    Just like `pdf_error` but over a list of ordernames.
    """
    if ordernames is None:
        ordernames = ["LO", "NLO", "aNNLO_PLUS_NNLL"]
    r_dl = dl
    for o in ordernames:
        r_dl = pdf_error(li, r_dl, o, confidence_level=confidence_level, n_jobs=n_jobs)

    return r_dl


def _pdf_error_single(members, i, dl, ordername, confidence_level=90):
    try:
        import lhapdf
    except ImportError:
        raise RuntimeError(
            "LHAPDF with python bindings needed to compute PDF uncertainties. Make sure you set the PYTHONPATH correctly (i.e. correct python version)."
        )
    if not lhapdf.availablePDFSets():
        raise RuntimeError(
            "No PDF sets found. Make sure the environment variable LHAPDF_DATA_DIR points to the correct location (.../share/LHAPDF)."
        )
    if dl["pdfset_nlo"][i] == 0 and dl["mu_f"][i] == 1.0 and dl["mu_r"][i] == 1.0:
        pdfset = lhapdf.getPDFSet(dl["pdf_nlo"][i])
        pdfs = [0.0] * pdfset.size
        ddl = dl[members].drop(columns=["pdfset_nlo", "precision", "max_iters"])
        bol = ddl.eq(ddl.iloc[i]).all(axis="columns")
        for j in range(len(dl["pdfset_nlo"])):
            if bol[j]:
                pdfs[dl["pdfset_nlo"][j]] = j

        # lo_unc = pdfset.uncertainty(
        #    [unv(dl["LO"][k]) for k in pdfs], -1)
        # if ordername == "LO":
        #    dl.loc[i, "LO_PDF_CENTRAL"] = unv(dl["LO"][i])
        #    dl.loc[i, "LO_PDF_ERRPLUS"] = 0.0
        #    dl.loc[i, "LO_PDF_ERRMINUS"] = 0.0
        #    dl.loc[i, "LO_PDF_ERRSYM"] = 0.0
        # else:
        # print([float(unv(dl[ordername][k])) for k in pdfs])
        nlo_unc = pdfset.uncertainty(
            [float(unv(dl[ordername][k])) for k in pdfs], confidence_level
        )
    return (i, nlo_unc)


def pdf_error(li, dl, ordername="LO", confidence_level=90, n_jobs=None):
    """
    Computes Parton Density Function (PDF) uncertainties through :func:`lhapdf.set.uncertainty`.

    Args:
        li (:obj:`list` of :class:`Input`): Input list.
        dl (:obj:`dict`): :class:`Result` dictionary with lists per entry.
        ordername (`str`): Name of the order.
        confidence_level (:obj:`double`): Confidence Level for PDF uncertainty

    Returns:
        :obj:`dict`: Modified `dl` with new `ordername_{PDF,PDF_CENTRAL,PDF_ERRPLUS,PDF_ERRMINUS,PDF_ERRSYM}` entries.
            - (`ordername`)_`PDF` contains a symmetrized :mod:`uncertainties` object.
    """
    global required_numerical_uncertainty_factor

    example = li[0]
    members = [
        attr
        for attr in dir(example)
        if not callable(getattr(example, attr)) and not attr.startswith("__")
    ]

    dl[ordername + "_PDF"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_PDF_CENTRAL"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_PDF_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_PDF_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_PDF_ERRSYM"] = np.array([None] * len(dl["pdfset_nlo"]))

    args = [
        {
            "members": members,
            "i": i,
            "dl": dl,
            "ordername": ordername,
            "confidence_level": confidence_level,
        }
        for i in range(len(dl["pdfset_nlo"]))
        if dl["pdfset_nlo"][i] == 0 and dl["mu_f"][i] == 1.0 and dl["mu_r"][i] == 1.0
    ]
    ret = tpqdm(
        args,
        _pdf_error_single,
        n_jobs=n_jobs if n_jobs is not None else mp.cpu_count(),
        argument_type="kwargs",
        desc="PDF uncertainty @ " + ordername,
    )
    for i, nlo_unc in ret:
        dl.loc[i, ordername + "_PDF_CENTRAL"] = nlo_unc.central
        dl.loc[i, ordername + "_PDF_ERRPLUS"] = nlo_unc.errplus
        dl.loc[i, ordername + "_PDF_ERRMINUS"] = -nlo_unc.errminus
        dl.loc[i, ordername + "_PDF_ERRSYM"] = nlo_unc.errsymm
        # TODO error sym to minus and plus
        # if :
        if ordername != "LO" and (
            usd(dl[ordername][i]) * required_numerical_uncertainty_factor
            > dl[ordername + "_PDF_ERRPLUS"][i] - dl[ordername + "_PDF_ERRMINUS"][i]
        ):
            rel = unv(dl[ordername][i])
            warnings.warn(
                "too bad numerical precision vs pdf @ "
                + ordername
                + " num: "
                + str(usd(dl[ordername][i]) / rel * 100.0)
                + "% vs "
                + str(dl[ordername + "_PDF_ERRPLUS"][i] / rel * 100.0)
                + "% to pdf: "
                + str(dl[ordername + "_PDF_ERRMINUS"][i] / rel * 100.0)
                + "%",
                RuntimeWarning,
            )

    mask = dl[ordername + "_PDF_CENTRAL"].notnull()
    dl.loc[mask, ordername + "_PDF"] = unp.uarray(
        unv(dl[ordername + "_PDF_CENTRAL"][mask])
        + dl[ordername + "_PDF_ERRPLUS"][mask] / 2.0
        + dl[ordername + "_PDF_ERRMINUS"][mask] / 2.0,
        (dl[ordername + "_PDF_ERRPLUS"][mask] - dl[ordername + "_PDF_ERRMINUS"][mask])
        / 2.0,
    )

    return dl


def scale_errors(li, dl, ordernames=None, n_jobs=None):
    """
    Just like `scale_error` but over a list of ordernames.
    """
    if ordernames is None:
        ordernames = ["LO", "NLO", "aNNLO_PLUS_NNLL"]
    r_dl = dl
    for o in ordernames:
        r_dl = scale_error(li, r_dl, o, n_jobs=n_jobs)

    return r_dl


def _scale_error_single(members, i, dl, ordername="LO"):
    if dl["pdfset_nlo"][i] == 0 and dl["mu_f"][i] == 1.0 and dl["mu_r"][i] == 1.0:
        scales = []
        ddl = dl[members].drop(columns=["mu_f", "mu_r", "precision", "max_iters"])
        bol = ddl.eq(ddl.iloc[i]).all(axis="columns")
        for j in range(len(dl["pdfset_nlo"])):
            if bol[j]:
                scales.append(j)
    # index, errplus,errminus
    return (
        i,
        np.max([unv(dl[ordername][k]) for k in scales]) - unv(dl[ordername][i]),
        np.min([unv(dl[ordername][k]) for k in scales]) - unv(dl[ordername][i]),
    )


def scale_error(li, dl, ordername="LO", n_jobs=None):
    """
    Computes seven-point scale uncertainties from the results where the renormalization and factorization scales are varied by factors of 2 and  relative factors of four are excluded (cf. :meth:`seven_point_scan`).

    Args:
        li (:obj:`list` of :class:`Input`): Input list.
        dl (:obj:`dict`): :class:`Result` dictionary with lists per entry.

    Returns:
        :obj:`dict`: Modified `dl` with new `ordername_{SCALE,SCALE_ERRPLUS,SCALE_ERRMINUS}` entries.
            - `ordername_SCALE` contains a symmetrized :mod:`uncertainties` object.
    """
    global required_numerical_uncertainty_factor
    example = li[0]
    members = [
        attr
        for attr in dir(example)
        if not callable(getattr(example, attr)) and not attr.startswith("__")
    ]
    dl[ordername + "_SCALE"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_SCALE_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_SCALE_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))

    args = [
        {
            "members": members,
            "i": i,
            "dl": dl,
            "ordername": ordername,
        }
        for i in range(len(dl["pdfset_nlo"]))
        if dl["pdfset_nlo"][i] == 0 and dl["mu_f"][i] == 1.0 and dl["mu_r"][i] == 1.0
    ]
    ret = tpqdm(
        args,
        _scale_error_single,
        n_jobs=n_jobs if n_jobs is not None else mp.cpu_count(),
        argument_type="kwargs",
        desc="Scale uncertainty @ " + ordername,
    )
    for i, errplus, errminus in ret:
        # lo_unc = pdfset.uncertainty(
        #    [unv(dl["LO"][k]) for k in pdfs], -1)
        dl.loc[i, ordername + "_SCALE_ERRPLUS"] = errplus
        dl.loc[i, ordername + "_SCALE_ERRMINUS"] = errminus
        if (
            usd(dl[ordername][i]) * required_numerical_uncertainty_factor
            > dl[ordername + "_SCALE_ERRPLUS"][i] - dl[ordername + "_SCALE_ERRMINUS"][i]
        ):
            rel = unv(dl[ordername][i])
            warnings.warn(
                "too bad numerical precision vs scale @ num:"
                + ordername
                + " "
                + str(usd(dl[ordername][i]) / rel * 100.0)
                + "% vs scale:"
                + str(dl[ordername + "_SCALE_ERRPLUS"][i] / rel * 100.0)
                + "% to "
                + str(dl[ordername + "_SCALE_ERRMINUS"][i] / rel * 100.0)
                + "%",
                RuntimeWarning,
            )

    mask = dl[ordername + "_SCALE_ERRPLUS"].notnull()
    dl.loc[mask, ordername + "_SCALE"] = unp.uarray(
        unv(dl[ordername][mask])
        + dl[ordername + "_SCALE_ERRPLUS"][mask] / 2.0
        + dl[ordername + "_SCALE_ERRMINUS"][mask] / 2.0,
        (
            +dl[ordername + "_SCALE_ERRPLUS"][mask]
            - dl[ordername + "_SCALE_ERRMINUS"][mask]
        )
        / 2.0,
    )

    return dl


def combine_errors(dl, ordernames=None):
    """
    Just like `combine_error` but over a list of ordernames.
    """
    if ordernames is None:
        ordernames = ["LO", "NLO", "aNNLO_PLUS_NNLL"]
    r_dl = dl
    for o in ordernames:
        r_dl = combine_error(r_dl, o)

    return r_dl


def combine_error(dl: dict, ordername="LO"):
    """
    Combines seven-point scale uncertainties and pdf uncertainties from the results by Pythagorean addition.

    Note:
        Running :func:`scale_errors` and :func:`pdf_errors` before is necessary.

    Args:
        dl (:obj:`dict`): :class:`Result` dictionary with lists per entry.

    Returns:
        :obj:`dict`: Modified `dl` with new `ordername_{COMBINED,ERRPLUS,ERRMINUS}` entries.
            - `ordername_COMBINED` contains a symmetrized :mod:`uncertainties` object.
    """
    dl[ordername + "_NOERR"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_ERRPLUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_ERRMINUS"] = np.array([None] * len(dl["pdfset_nlo"]))
    dl[ordername + "_COMBINED"] = np.array([None] * len(dl["pdfset_nlo"]))

    mask = dl[ordername + "_PDF_CENTRAL"].notnull()
    dl.loc[mask, ordername + "_NOERR"] = unv(dl[ordername + ""][mask]).astype(float)
    dl.loc[mask, ordername + "_ERRPLUS"] = np.sqrt(
        dl[ordername + "_PDF_ERRPLUS"][mask].astype(float) ** 2
        + dl[ordername + "_SCALE_ERRPLUS"][mask].astype(float) ** 2
    )
    dl.loc[mask, ordername + "_ERRMINUS"] = -np.sqrt(
        dl[ordername + "_PDF_ERRMINUS"][mask].astype(float) ** 2
        + dl[ordername + "_SCALE_ERRMINUS"][mask].astype(float) ** 2
    )
    dl.loc[mask, ordername + "_COMBINED"] = unp.uarray(
        unv(dl[ordername + ""][mask])
        + dl[ordername + "_ERRPLUS"][mask] / 2.0
        + dl[ordername + "_ERRMINUS"][mask] / 2.0,
        (+dl[ordername + "_ERRPLUS"][mask] - dl[ordername + "_ERRMINUS"][mask]) / 2.0,
    )

    return dl

def asym_to_sym_error(central,errminus,errplus):
    return ufloat(
        central + (errplus + errminus) / 2.0,
        (errplus - errminus) / 2.0,
    )

def add_errors(error1,error2):
    return (error1**2 + error2**2)**0.5

def asym_to_sym_combined_error(central,errminus1,errplus1,errminus2,errplus2):
    return asym_to_sym_error(central,add_errors(errplus1,errplus2),add_errors(errminus1,errminus2))


import importlib.resources
import os


def list_files():
    """
    List all files in the data directory

    Returns
    -------
    list
            List of files in the data directory

    """
    return [
        os.path.basename(b)
        for b in importlib.resources.files(".".join(__name__.split(".")[:-1]))
        .joinpath("json")
        .iterdir()
    ]


def get_file(filename):
    """
    Get the content of a file in the data directory

    Parameters
    ----------
    filename : str
            Name of the file

    Returns
    -------
    str
            Content of the file

    """
    return (
        importlib.resources.files(".".join(__name__.split(".")[:-1]))
        .joinpath("json")
        .joinpath(filename)
    )



"""Runs MadGraph."""
import os
import pkgutil
import subprocess
import time
from string import Template
from typing import List

from hepi.input import Order
from hepi.run import Runner, RunParam

from .. import Input, Result, get_output_dir
from .result import is_valid, parse_single


class MadGraphRunParams(RunParam):
    """Parameters for MadGraph."""

    def __init__(self, dic, skip=False, madstr=True):
        super().__init__(skip)
        self.dic = dic
        # self.skip = skip
        self.madstr = madstr


class MadGraphRunner(Runner):
    def orders(self) -> List[Order]:
        return [Order.LO, Order.NLO]

    def _check_path(self) -> bool:
        if os.path.exists(os.path.expanduser(self.get_path() + "/bin/mg5_aMC")):
            self.set_path(self.get_path() + "/bin/mg5_aMC")
            return True
        if self.get_path().endswith("mg5_aMC"):
            return True
        return False

    def _check_input(self, param: Input, **kwargs) -> bool:
        """Checks input parameter for compatibility with Prospino"""
        return True

    def _is_valid(self, file: str, p: Input, d) -> bool:
        if not super()._is_valid(file, p, d):
            return False
        return is_valid(file, p, d)

    def _parse_file(self, file: str) -> Result:
        return parse_single(file)

    def _run(self, rps: List[RunParam], wait=True, parallel=True, sleep=0, **kwargs):
        # TODO clean up on exit emergency
        skipall = True
        for rp in rps:
            if not rp.skip:
                skipall = False
        template = ""
        if not skipall:
            lo = (
                "&& nice -n 5 {dir}/bin/calculate_xsect LO -f >> {out} "
                if rps[0].dic["order"] == Order.NLO
                else ""
            )
            template = (
                "rm -rf {dir} && cp -r "
                + rps[0].dic["bdir"]
                + ' {dir}  && cp {slha} {dir}/Cards/param_card.dat && cp {run} {dir}/Cards/run_card.dat && echo "nb_core = 1" >> {dir}/Cards/amcatnlo_configuration.txt '
                + lo
                + "&& nice -n 5 {dir}/bin/calculate_xsect -f >> {out}  && rm -rf {dir}"
            )
            print(rps[0].dic["out"])
        if not rps[0].skip:
            mgcom = ""
            if rps[0].madstr:
                mgcom = ' --mode="MadSTR"'
            com = (
                self.get_path()
                + mgcom
                + " --file {in} >> {out} && cp {slha} {bdir}/Cards/param_card.dat && cp {run} {bdir}/Cards/run_card.dat && sed -i 's/.*= req_acc_FO/ 1 = req_acc_FO/g' {bdir}/Cards/run_card.dat && echo \"automatic_html_opening = False\" >> {bdir}/Cards/amcatnlo_configuration.txt && nice -n 5 {bdir}/bin/calculate_xsect -f"
            )
            pp = subprocess.Popen(com.format(**rps[0].dic), shell=True)
            pp.wait()
        # Run commands in parallel
        processes = []

        for rp in rps:
            if not rp.skip:
                command = template.format(**rp.dic)
                # print(command)
                process = subprocess.Popen(command, shell=True)
                processes.append(process)
                if not parallel:
                    process.wait()
                # Forced delay to prevent overloading clusters when registering jobs
                time.sleep(sleep)
        if wait:
            # Collect statuses
            output = [p.wait() for p in processes]
            return output
        return []

    def _prepare(self, p: Input, **kwargs) -> RunParam:
        rp = super()._prepare(p, **kwargs)
        name = rp.name
        if not rp.skip:
            d = p.__dict__
            d["dir"] = get_output_dir() + name + ".dir"
            d["bdir"] = get_output_dir() + name + ".bdir"

            infile = ""
            if p.order is Order.LO:
                infile = "lo.mg"
            elif p.order is Order.NLO:
                infile = "nlo.mg"
            else:
                raise ValueError("Madgraph only supported for LO/NLO.")

            data = pkgutil.get_data(__name__, infile).decode("utf-8")
            src = Template(data)
            result = src.substitute(d)
            with open(get_output_dir() + name + ".mg", "w") as tmp:
                tmp.write(result)
            if not rp.skip:
                with open(get_output_dir() + name + ".out", "w") as tmp:
                    tmp.write(result + "\n\n")

            if p.order == Order.LO:
                mgfile = "run_card_no_madstr.dat"
            elif p.order == Order.NLO:
                mgfile = "run_card_with_madstr.dat"
            else:
                raise ValueError("Order must be one of LO/NLO in MadGraph.")

            if "madstr" in kwargs and not kwargs["madstr"]:
                mgfile = "run_card_no_madstr.dat"
                rp.madstr = False
            else:
                rp.madstr = True
            data = pkgutil.get_data(__name__, mgfile).decode("utf-8")

            src = Template(data)
            result = src.substitute(d)
            with open(get_output_dir() + name + ".dat", "w") as tmp:
                tmp.write(result)
            if not rp.skip:
                with open(get_output_dir() + name + ".out", "a") as tmp:
                    tmp.write(result + "\n\n")

            sname = d["slha"]
            with open(get_output_dir() + sname, "r") as f:
                # src = Template(f.read())
                # result = src.substitute(d)
                # open(get_input_dir() + sname + ".in", "w").write(result)
                if not rp.skip:
                    with open(get_output_dir() + name + ".out", "a") as a:
                        a.write(f.read() + "\n\n")
            rp.dic = {
                "in": get_output_dir() + name + ".mg",
                "dir": d["dir"],
                "bdir": d["bdir"],
                "run": get_output_dir() + name + ".dat",
                "slha": get_output_dir() + sname,
                "out": get_output_dir() + name + ".out",
                "order": p.order,
            }
            rp.out_file = rp.dic["out"]
        return rp


# Legacy
default_madgraph_runner = MadGraphRunner("/opt/MG5_aMC_v2_7_0/")
"""Default MadGraph Runner to provide backward compatibility"""
run = default_madgraph_runner.run
set_path = default_madgraph_runner.set_path
get_path = default_madgraph_runner.get_path

# print(sp)
# names = queue(sp)
# run(names)



import os
import pkgutil
import re
import shutil
import warnings
from string import Template

from uncertainties import ufloat_fromstr

from hepi.input import Order

from .. import Input, Result, get_output_dir


class ResumminoResult(Result):
    """
    Resummino Result Data.

    Attributes:
        VNLO (double): virtual NLO result.
        RNLO (double): real NLO result.
        P_PLUS_K (double): collineare remainders P+K NLO result.
        RNLOG (double): real NLO gluon result.
        RNLOQ (double): real NLO quark result.
        VNLO_PLUS_P_PLUS_K (double): VNLO+P+K result.
        RNLO_PLUS_VNLO_PLUS_P_PLUS_K (double): RNLO+VNLO+P+K result.
    """

    def __init__(
        self, lo, nlo, nlo_plus_nll, annlo_plus_nnll, vnlo, p_plus_k, rnlog, rnloq
    ):
        """
        Sets given and computes dependent ``Attributes``.

        Args:
            lo (:obj:`double`): corresponds to :attr:`LO`.
            nlo (:obj:`double`): corresponds to :attr:`NLO`.
            nlo_plus_nll (:obj:`double`): corresponds to :attr:`NLO_PLUS_NLL`.
            vnlo (:obj:`double`): corresponds to :attr:`VNLO`.
            p_plus_k (:obj:`double`): corresponds to :attr:`P_PLUS_K`.
            rnlog (:obj:`double`): corresponds to :attr:`RNLOG`.
            rnloq (:obj:`double`): corresponds to :attr:`RNLOQ`.
        """
        Result.__init__(self, lo, nlo, nlo_plus_nll, annlo_plus_nnll)
        self.VNLO = vnlo
        self.P_PLUS_K = p_plus_k
        self.RNLOG = rnlog
        self.RNLOQ = rnloq
        if not (vnlo is None or p_plus_k is None):
            self.VNLO_PLUS_P_PLUS_K = self.VNLO + self.P_PLUS_K
        else:
            self.VNLO_PLUS_P_PLUS_K = None
        if not (rnlog is None or rnloq is None):
            self.RNLO = rnlog + rnloq
        else:
            self.RNLO = None
        if not (self.RNLO is None or self.VNLO_PLUS_P_PLUS_K is None):
            self.RNLO_PLUS_VNLO_PLUS_P_PLUS_K = self.RNLO + self.VNLO_PLUS_P_PLUS_K
        else:
            self.RNLO_PLUS_VNLO_PLUS_P_PLUS_K = None


def is_valid(file: str, p: Input, d, **kwargs) -> bool:
    """
    Verifies that an file is a complete output.

    Args:
        file (str): File path to be parsed.
        p (:class:`hepi.Input`): Onput parameters.
        d (:obj:`dict`): Param dictionary.

    Returns:
        bool : True if `file` could be parsed.
    """
    order = p.order
    data = pkgutil.get_data(__name__, "plot_template.in").decode("utf-8")
    src = Template(data)
    result = src.substitute(d)
    sname = d["slha"]
    with open(file, mode="r") as f:
        with open(get_output_dir() + sname, "r") as sf:
            if not kwargs["ignore_error"] and not f.read().startswith(
                result + "\n\n" + sf.read()
            ):
                warnings.warn(
                    "Possible hash collision in "
                    + file
                    + " -> moved to "
                    + file
                    + ".old",
                    RuntimeWarning,
                )
                shutil.move(file, file + ".old")
                return False
    res = parse_single(file)
    if res.LO is not None and order is Order.LO:
        return True
    if res.LO is not None and res.NLO is not None and order is Order.NLO:
        return True
    if (
        res.LO is not None
        and res.NLO is not None
        and res.NLO_PLUS_NLL is not None
        and order is Order.NLO_PLUS_NLL
    ):
        return True
    if (
        res.LO is not None
        and res.NLO is not None
        and res.aNNLO_PLUS_NNLL is not None
        and order is Order.aNNLO_PLUS_NNLL
    ):
        return True
    print("RESTART", res.LO, res.NLO, res.NLO_PLUS_NLL, file)
    return False


def parse_single(file: str) -> ResumminoResult:
    """
    Extracts LO, NLO and NLO+NLL from Resummino output file.

    Args:
        file (str): File path to be parsed.

    Returns:
        :class:`ResumminoResult` : If a value is not found in the file None is used.

    """
    # TODO generalize units like RS
    lo_pattern = re.compile(r"^LO = \((.*)\) pb")
    nlo_pattern = re.compile(r"^NLO = \((.*)\) pb")
    nll_pattern = re.compile(r"^NLO\+NLL = \((.*)\) pb")
    nnll_pattern = re.compile(r"^aNNLO\+NNLL = \((.*)\) pb")

    vnlo_pattern = re.compile(r"^vNLO = \((.*)\) pb")
    ppk_pattern = re.compile(r"^P\+K = \((.*)\) pb")
    rnlog_pattern = re.compile(r"^rNLOg = \((.*)\) pb")
    rnloq_pattern = re.compile(r"^rNLOq = \((.*)\) pb")

    lo_result = None
    nlo_result = None
    nll_result = None
    nnll_result = None

    vnlo_result = None
    ppk_result = None
    rnlog_result = None
    rnloq_result = None
    with open(file) as output:
        for line in output:
            tmp = lo_pattern.search(line)
            if tmp is not None:
                lo_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))
            tmp = nlo_pattern.search(line)
            if tmp is not None:
                nlo_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))
            tmp = nll_pattern.search(line)
            if tmp is not None:
                nll_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))
            tmp = nnll_pattern.search(line)
            if tmp is not None:
                nnll_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))
            tmp = vnlo_pattern.search(line)
            if tmp is not None:
                vnlo_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))
            tmp = ppk_pattern.search(line)
            if tmp is not None:
                ppk_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))
            tmp = rnlog_pattern.search(line)
            if tmp is not None:
                rnlog_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))
            tmp = rnloq_pattern.search(line)
            if tmp is not None:
                rnloq_result = ufloat_fromstr(tmp.group(1).replace("+-", "+/-"))
    if lo_result is not None and lo_result.nominal_value < 0:
        warnings.warn("LO < 0")
    if nlo_result is not None and nlo_result.nominal_value < 0:
        warnings.warn("NLO < 0")
    if nll_result is not None and nll_result.nominal_value < 0:
        warnings.warn("NLL < 0")
    if nnll_result is not None and nnll_result.nominal_value < 0:
        warnings.warn("NNLL < 0")
    return ResumminoResult(
        lo_result,
        nlo_result,
        nll_result,
        nnll_result,
        vnlo_result,
        ppk_result,
        rnlog_result,
        rnloq_result,
    )



from hepi.results import Result, add_errors, asym_to_sym_error


class NLLFastResult(Result):
    """
    (N)NLL-fast Result Data.
    """

    def __init__(
        self,
        LO,
        NLO,
        NLO_PDF_ERRMINUS,
        NLO_PDF_ERRPLUS,
        NLO_PLUS_NLL,
        NLO_PLUS_NLL_SCALE_ERRMINUS,
        NLO_PLUS_NLL_SCALE_ERRPLUS,
        NLO_PLUS_NLL_PDF_ERRMINUS,
        NLO_PLUS_NLL_PDF_ERRPLUS,
    ):
        Result.__init__(self, LO, NLO, NLO_PLUS_NLL)
        self.LO_NOERR= LO
        self.NLO_NOERR= NLO
        self.NLO_PDF_ERRMINUS=NLO_PDF_ERRMINUS
        self.NLO_PDF_ERRPLUS =NLO_PDF_ERRPLUS
        self.NLO_PDF = asym_to_sym_error(NLO, NLO_PDF_ERRMINUS, NLO_PDF_ERRPLUS)

        self.NLO_PLUS_NLL_NOERR = NLO_PLUS_NLL

        self.NLO_PLUS_NLL_SCALE_ERRMINUS=NLO_PLUS_NLL_SCALE_ERRMINUS
        self.NLO_PLUS_NLL_SCALE_ERRPLUS=NLO_PLUS_NLL_SCALE_ERRPLUS
        self.NLO_PLUS_NLL_SCALE=asym_to_sym_error(NLO_PLUS_NLL, NLO_PLUS_NLL_SCALE_ERRMINUS, NLO_PLUS_NLL_SCALE_ERRPLUS)

        self.NLO_PLUS_NLL_PDF_ERRMINUS=NLO_PLUS_NLL_PDF_ERRMINUS
        self.NLO_PLUS_NLL_PDF_ERRPLUS=NLO_PLUS_NLL_PDF_ERRPLUS
        self.NLO_PLUS_NLL_PDF=asym_to_sym_error(NLO_PLUS_NLL, NLO_PLUS_NLL_PDF_ERRMINUS, NLO_PLUS_NLL_PDF_ERRPLUS)

        self.NLO_PLUS_NLL_ERRMINUS=-add_errors(NLO_PLUS_NLL_PDF_ERRMINUS, NLO_PLUS_NLL_SCALE_ERRMINUS)
        self.NLO_PLUS_NLL_ERRPLUS=add_errors(NLO_PLUS_NLL_PDF_ERRPLUS, NLO_PLUS_NLL_SCALE_ERRPLUS)
        self.NLO_PLUS_NLL = asym_to_sym_error(NLO_PLUS_NLL, self.NLO_PLUS_NLL_ERRMINUS, self.NLO_PLUS_NLL_ERRPLUS)





import os
import subprocess
import warnings
from typing import List

from hepi.input import Input, update_slha
from hepi.run import Runner


class SoftsusyRunner(Runner):
    def _check_path(self) -> bool:
        if os.path.exists(os.path.expanduser(self.get_path() + "/bin/softpoint.x")):
            self.set_path(self.get_path() + "/bin/softpoint.x")
            return True
        if self.get_path().endswith("softpoint.x"):
            return True
        return False

    def run(self, slhas: List[Input], **kwargs) -> List[Input]:
        """
        Run the passed list of parameters for softsusy.

        Args:
            slhas (:obj:`list` of :class:`Input`): Input parameters with a SLHA file that can be processed by softsusy.
        Returns:
            :obj:`list` of :class:`Input`
        """
        if not self._check_path():
            warnings.warn("The path is not valid for " + self.get_name())
        for s in slhas:
            # Remove Creation time for hash-/caching
            comm = (
                "cp "
                + self.get_output_dir()
                + s.slha
                + " softsusy_tmp.in && "
                + get_path()
                + " leshouches < softsusy_tmp.in > Softsusy.spc"
                + " && mv "
                + "Softsusy.spc "
                + self.get_output_dir()
                + s.slha
                #                + " && sed -i '/Created/d' "
                #                + self.get_output_dir()
                #                + s.slha
            )
            # print(comm)
            with subprocess.Popen(
                comm, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            ) as proc:
                proc.wait()
                # get proc return code
                if proc.returncode != 0:
                    warnings.warn("softsusy failed")
                update_slha(s)
        return slhas


# Backward compatibility
spheno_default_runner = SoftsusyRunner("softpoint.x")
"""Default SoftSusy Runner to provide backward compatibility"""
run = spheno_default_runner.run
set_path = spheno_default_runner.set_path
get_path = spheno_default_runner.get_path



from collections.abc import Iterable

import matplotlib as mpl
import matplotlib.cm as cm
import matplotlib.pyplot as plt
import numpy as np
import pyslha
from matplotlib import colors
from matplotlib.image import NonUniformImage
from matplotlib.ticker import NullFormatter
from scipy import integrate as auc
from scipy.interpolate import BSpline, make_interp_spline
from smpl import plot as splot

from ..input import Input, get_output_dir, replace_macros
from ..particles import get_name


def title(
    i: Input,
    axe=None,
    scenario=None,
    diff_L_R=None,
    extra="",
    cms_energy=True,
    pdf_info=True,
    id=False,
    **kwargs
):
    """Sets the title on axis `axe`."""
    if axe is None:
        axe = plt.gca()
    axe.set_title(
        "$pp\\to"
        + get_name(i.particle1)
        + get_name(i.particle2)
        + "$"
        + (" at $\\sqrt{S} = " + str(i.energy / 1000) + "$ TeV" if cms_energy else "")
        + ((" for " + i.slha.split(".")[0]) if scenario is None else (" " +scenario))
        + (" with " + i.pdf_nlo if pdf_info else "")
        + " "
        + extra
        + ((" [" + i.id + "]") if id else "")
    )


def energy_plot(
    dict_list,
    y,
    yscale=1.0,
    xaxis="E [GeV]",
    yaxis="$\\sigma$ [pb]",
    label=None,
    **kwargs
):
    """Plot energy on the x-axis."""
    plot(
        dict_list,
        "energy",
        y,
        label=label,
        xaxis=xaxis,
        yaxis=yaxis,
        logy=True,
        yscale=yscale,
        **kwargs
    )


def combined_mass_plot(dict_list, y, part, label=None, **kwargs):
    if label == "":
        kwargs["label"] = ""
    mass_plot(
        dict_list,
        y + "_NOERR",
        part,
        next_color=False,
        fmt=" ",
        fill=False,
        interpolate_label=label,
        **kwargs
    )
    if "label" in kwargs:
        del kwargs["label"]
    mass_plot(
        dict_list,
        y + "_COMBINED",
        part,
        next_color=True,
        fit_fmt=" ",
        fill=True,
        label="",
        interpolate_lower_uncertainty=False,
        **kwargs
    )


def combined_plot(dict_list, x, y, label=None, **kwargs):
    if label == "":
        kwargs["label"] = ""
    plot(
        dict_list,
        x,
        y + "_NOERR",
        next_color=False,
        fmt=" ",
        fill=False,
        interpolate_label=label,
        **kwargs
    )
    if "label" in kwargs:
        del kwargs["label"]
    plot(
        dict_list,
        x,
        y + "_COMBINED",
        next_color=True,
        fit_fmt=" ",
        fill=True,
        label="",
        interpolate_lower_uncertainty=False,
        **kwargs
    )


def mass_plot(
    dict_list,
    y,
    part,
    logy=True,
    yaxis="$\\sigma$ [pb]",
    yscale=1.0,
    label=None,
    xaxis=None,
    **kwargs
):
    if not "mass_" + str(part) in dict_list:
        dict_list["mass_" + str(part)] = get_mass(dict_list, abs(part))
    if xaxis is None:
        xaxis = "$m_{" + get_name(part) + "}$ [GeV]"
    plot(
        dict_list,
        "mass_" + str(part),
        y,
        label=label,
        yaxis=yaxis,
        xaxis=xaxis,
        logy=logy,
        yscale=yscale,
        **kwargs
    )


def mass_vplot(
    dict_list,
    y,
    part,
    logy=True,
    yaxis="$\\sigma$ [pb]",
    yscale=1.0,
    label=None,
    mask=None,
    **kwargs
):
    vplot(
        get_mass(dict_list, part)[mask],
        y[mask],
        label=label,
        xaxis="$m_{" + get_name(part) + "}$ [GeV]",
        yaxis=yaxis,
        logy=logy,
        yscale=yscale,
        mask=mask,
        **kwargs
    )


def get_mass(l: dict, iid: int):
    """
    Get the mass of particle with id `iid` out of the list in the "slha" element in the dict.

    Returns
        :obj:`list` of float : masses of particles in each element of the dict list.

    """
    ret = []
    for s in l["slha"]:
        d = pyslha.read(get_output_dir() + s)
        ret.append(d.blocks["MASS"][abs(iid)])

    return np.array(ret)


def plot(
    dict_list,
    x,
    y,
    label=None,
    xaxis="M [GeV]",
    yaxis="$\\sigma$ [pb]",
    ratio=False,
    K=False,
    K_plus_1=False,
    logy=True,
    yscale=1.0,
    mask=None,
    **kwargs
) -> None:
    """
    Creates a plot based on the entries `x`and `y` in `dict_list`.

    Examples

    .. plot::
        :include-source:

        >>> import urllib.request
        >>> import hepi
        >>> dl = hepi.load(urllib.request.urlopen(
        ... "https://raw.githubusercontent.com/fuenfundachtzig/xsec/master/json/pp13_hino_NLO%2BNLL.json"
        ... ))
        >>> hepi.plot(dl,"N1","NLO_PLUS_NLL_COMBINED",xaxis="$m_{\\\\tilde{\\\\chi}_1^0}$ [GeV]")
    """
    if isinstance(y, Iterable) and not isinstance(y, str):
        for yi in y:
            plot(
                dict_list,
                x,
                yi,
                label,
                xaxis,
                yaxis,
                ratio,
                K,
                K_plus_1,
                logy,
                yscale,
                mask,
                **kwargs
            )
        return
    # TODO use kwargs
    if label is None:
        label = y.replace("_PLUS_", "+").replace("_OVER_", "/")
    if label == "":
        label = None

    vx = dict_list[x].to_numpy()[mask]
    vy = dict_list[y].to_numpy()[mask]

    if K:
        yaxis = "$K$"
        yscale = 1.0
        vy = vy / splot.unv(dict_list["LO"].to_numpy()[mask])
        if K_plus_1:
            vy = vy + vy / vy
    if ratio:
        yaxis = "Ratio"
        yscale = 1.0
        vy = vy / splot.unv(vy)

    vplot(vx, vy, label, xaxis, yaxis, logy, yscale, mask=mask, **kwargs)


def index_open(var, idx):
    if len(idx) == 1:
        return var[idx]
    return index_open(var[idx[0]], idx[1:])


def slha_data(li, index_list):
    vx = []
    for l in li:
        b = pyslha.read(get_output_dir() + l.slha)
        vx.append(index_open(b.blocks, index_list))
    return np.array(vx)


def slha_plot(li, x, y, **kwargs):
    vx = slha_data(li, x)
    vy = slha_data(li, y)

    vplot(np.array(vx), np.array(vy), **kwargs)


def vplot(
    x,
    y,
    label=None,
    xaxis="E [GeV]",
    yaxis="$\\sigma$ [pb]",
    logy=True,
    yscale=1.0,
    interpolate=True,
    plot_data=True,
    data_color=None,
    mask=-1,
    fill=False,
    data_fmt=".",
    fmt="-",
    print_area=False,
    sort=True,
    **kwargs
):
    """
    Creates a plot based on the values in `x`and `y`.

    """
    color = data_color
    if label is None:
        # label = "??"
        pass
    if mask is None:
        x = x[0]
        y = y[0]
    if sort:
        permute = x.argsort(kind="stable")
        vx = x[permute]
        vy = y[permute]
    else:
        vx = x
        vy = y

    if data_color is None and "next_color" not in kwargs:
        if "axes" in kwargs and kwargs["axes"] is not None:
            (bl,) = kwargs["axes"].plot([], [])
        else:
            (bl,) = plt.gca().plot([], [])
        color = bl.get_color()
    iii = splot.data(
        vx,
        vy * yscale,
        label=label,
        xlabel=xaxis,
        ylabel=yaxis,
        logy=logy,
        data_color=color,
        fit_color=color,
        also_data=plot_data,
        interpolate=interpolate,
        fmt=data_fmt,
        sigmas=0 if not fill else 1,
        **kwargs
    )
    if iii is not None:
        # ii = iii[0]
        ix = iii[1]
        iy = iii[2]
    if print_area:
        print(
            "computed AUC using scipy.integrate.simpson: {}".format(auc.simpson(iy, ix))
        )
    if (np.any(np.less(vy, 0)) or (interpolate and np.any(np.less(iy, 0)))) and logy:
        splot.data(
            vx,
            -vy * yscale,
            label="-" + label,
            xaxis=xaxis,
            yaxis=yaxis,
            logy=logy,
            data_color=color,
            fit_color=color,
            also_data=plot_data,
            interpolate=interpolate,
            interpolate_fmt=(0, (3, 1, 3, 1, 1, 1)),
            fmt=data_fmt,
            sigmas=0 if not fill else 1,
            **kwargs
        )


def mass_mapplot(
    dict_list,
    part1,
    part2,
    z,
    logz=True,
    zaxis="$\\sigma$ [pb]",
    zscale=1.0,
    label=None,
):
    mapplot(
        dict_list,
        "mass_" + str(part1),
        "mass_" + str(part2),
        z,
        xaxis="$M_{" + get_name(part1) + "}$ [GeV]",
        yaxis="$M_{" + get_name(part2) + "}$ [GeV]",
        zaxis=zaxis,
        logz=logz,
        zscale=zscale,
    )


def mapplot(dict_list, x, y, z, xaxis=None, yaxis=None, zaxis=None, **kwargs):
    """

    Examples

    .. plot::
        :include-source:

        >>> import urllib.request
        >>> import hepi

        >>> dl = hepi.load(urllib.request.urlopen(
        ... "https://raw.githubusercontent.com/APN-Pucky/xsec/master/json/pp13_SGmodel_GGxsec_NLO%2BNLL.json"
        ... ),dimensions=2)
        >>> hepi.mapplot(dl,"gl","sq","NLO_PLUS_NLL_COMBINED",xaxis="$m_{\\\\tilde{g}}$ [GeV]",yaxis="$m_{\\\\tilde{q}}$ [GeV]" , zaxis="$\\\\sigma_{\\\\mathrm{NLO+NLL}}$ [pb]")
    """
    if xaxis is None:
        xaxis = x
    if yaxis is None:
        yaxis = y
    if zaxis is None:
        zaxis = replace_macros(z)
    vx = dict_list[x]
    vy = dict_list[y]
    vz = dict_list[z]
    splot.plot2d(
        vx, vy, vz, style="pcolormesh", xlabel=xaxis, ylabel=yaxis, zlabel=zaxis, **kwargs
    )


map_vplot = lambda *a, **da: splot.plot2d(*a, style="pcolormesh", **da)
scatter_vplot = lambda *a, **da: splot.plot2d(*a, style="scatter", **da)


def scatterplot(dict_list, x, y, z, xaxis=None, yaxis=None, zaxis=None, **kwargs):
    """
    Scatter map 2d.
    Central color is the central value, while the inner and outer ring are lower and upper bounds of the uncertainty interval.

    Examples

    .. plot::
        :include-source:

        >>> import urllib.request
        >>> import hepi
        >>> dl = hepi.load(urllib.request.urlopen(
        ... "https://raw.githubusercontent.com/APN-Pucky/xsec/master/json/pp13_hinosplit_N2N1_NLO%2BNLL.json"
        ... ),dimensions=2)
        >>> hepi.scatterplot(dl,"N1","N2","NLO_PLUS_NLL_COMBINED",xaxis="$m_{\\\\tilde{\\\\chi}_1^0}$ [GeV]",yaxis="$m_{\\\\tilde{\\\\chi}_2^0}$ [GeV]" , zaxis="$\\\\sigma_{\\\\mathrm{NLO+NLL}}$ [pb]")

    """
    if xaxis is None:
        xaxis = x
    if yaxis is None:
        yaxis = y
    if zaxis is None:
        zaxis = replace_macros(z)
    vx = dict_list[x]
    vy = dict_list[y]
    vz = dict_list[z]
    splot.plot2d(
        vx, vy, vz, style="scatter", xlabel=xaxis, ylabel=yaxis, zlabel=zaxis, **kwargs
    )


fig = None
axs = None
lines = []
labels = []


def err_plt(axes, x, y, label=None, error=False):
    v = label
    ind = np.argsort(splot.unv(x), kind="stable")
    if error:
        l, _, _ = axes.errorbar(
            x.to_numpy()[ind], splot.unv(y)[ind], yerr=splot.usd(y), capsize=5, label=v
        )
        return l
    else:
        l = axes.plot(x.to_numpy()[ind], splot.unv(y)[ind], label=v)
        return l[0]


def scale_plot(
    dict_list,
    vl,
    seven_point_band=False,
    cont=False,
    error=True,
    li=None,
    plehn_color=False,
    yscale=1.0,
    unit="pb",
    yaxis=None,
    **kwargs
):
    """Creates a scale variance plot with 5 panels (xline)."""
    global fig, axs, lines, labels
    cycle_safe = mpl.rcParams["axes.prop_cycle"]
    if plehn_color:
        mpl.rcParams["axes.prop_cycle"] = mpl.cycler(color=["b", "r", "g"])
    if not cont:
        fig, axs = plt.subplots(1, 5, figsize=(12, 3), sharey=True)
        # Remove horizontal space between axes
        fig.subplots_adjust(wspace=0)
        if li is not None:
            title(axs[2], li[0], **kwargs)

    mr = dict_list["mu_r"]
    mf = dict_list["mu_f"]

    if not cont:
        lines = []
        labels = []
        axs[0].plot([], [], " ", color="k", label="$\\mu_R=" + "\\mu_F$")
        axs[1].plot(
            [], [], " ", color="k", label="$\\mu_R=" + str(np.max(mr)) + "\\mu_0$"
        )
        axs[2].plot(
            [], [], " ", color="k", label="$\\mu_F=" + str(np.min(mf)) + "\\mu_0$"
        )
        axs[3].plot(
            [], [], " ", color="k", label="$\\mu_R=" + str(np.min(mr)) + "\\mu_0$"
        )
        axs[4].plot(
            [], [], " ", color="k", label="$\\mu_F=" + str(np.max(mf)) + "\\mu_0$"
        )

    for v in vl:
        mv = dict_list[v] * yscale
        if seven_point_band:
            # mask = (mf/mr < 4.) & (mf/mr > 1./4.) & (mf <= 2) & (mf >= 1./2.) & (mr <= 2) & (mr >= 1./2.)
            mask = (
                ((mf == 2.0) & (mr == 2.0))
                | ((mf == 2.0) & (mr == 1.0))
                | ((mf == 1.0) & (mr == 1.0))
                | ((mf == 1.0) & (mr == 2.0))
                | ((mf == 1 / 2.0) & (mr == 1 / 2.0))
                | ((mf == 1 / 2.0) & (mr == 1.0))
                | ((mf == 1.0) & (mr == 1 / 2.0))
            )
            mvmax = np.max(splot.unv(mv[mask]))
            mvmin = np.min(splot.unv(mv[mask]))

        mask = mf == mr
        l = err_plt(axs[0], mf[mask], mv[mask], error=error)
        # l, _, _ = axs[0].errorbar(mf[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5, label=v)
        if seven_point_band:
            axs[0].fill_between(
                mf[mask], mvmax, mvmin, facecolor=l.get_color(), alpha=0.3
            )

        mask = mr == np.max(mr)
        l = err_plt(axs[1], mf[mask], mv[mask], error=error)
        lines.append(l)
        labels.append(
            "$\\sigma_{\\mathrm{"
            + v.replace("_PLUS_", "+").replace(" ", "\\ ")
            + "} }$"
        )
        # l, _, _ = axs[1].errorbar(mf[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5)
        if seven_point_band:
            axs[1].fill_between(
                mf[mask], mvmax, mvmin, facecolor=l.get_color(), alpha=0.3
            )

        mask = mf == np.min(mf)
        l = err_plt(axs[2], mr[mask], mv[mask], error=error)
        # l, _, _ = axs[2].errorbar(mr[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5)
        if seven_point_band:
            axs[2].fill_between(
                mr[mask], mvmax, mvmin, facecolor=l.get_color(), alpha=0.3
            )

        mask = mr == np.min(mr)
        l = err_plt(axs[3], mf[mask], mv[mask], error=error)
        # l, _, _ = axs[3].errorbar(mf[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5)
        if seven_point_band:
            axs[3].fill_between(
                mf[mask], mvmax, mvmin, facecolor=l.get_color(), alpha=0.3
            )

        mask = mf == np.max(mf)
        l = err_plt(axs[4], mr[mask], mv[mask], error=error)
        # l, _, _ = axs[4].errorbar(mr[mask], splot.unv(mv[mask]),
        #                          yerr=splot.usd(mv[mask]), capsize=5)
        if seven_point_band:
            f = axs[4].fill_between(
                mr[mask], mvmax, mvmin, facecolor=l.get_color(), alpha=0.3
            )
            lines.append(f)
            labels.append(
                "$\\Delta \\sigma_{\\mathrm{"
                + v.replace("NLO_PLUS_NLL", "NLO+NLL").replace(" ", "\\<space>")
                + "} }$"
            )

    axs[0].set_ylabel("$\\sigma$ [" + unit + "]" if yaxis is None else yaxis)

    axs[0].set_xscale("log")
    axs[0].set_xlim(np.min(mf), np.max(mf))
    axs[0].set_xlabel("$\\mu_{R,F}/\\mu_0$")

    # axs[1].plot(t, s2)
    axs[1].set_xscale("log")
    axs[1].set_xlim(np.max(mf), np.min(mf))
    axs[1].set_xticks([1.0])
    axs[1].xaxis.set_minor_formatter(NullFormatter())
    axs[1].set_xlabel("$\\mu_{F}/\\mu_0$")

    # axs[2].plot(t, s3)
    axs[2].set_xscale("log")
    axs[2].set_xlim(np.max(mf), np.min(mf))
    axs[2].set_xticks([1.0])
    axs[2].xaxis.set_minor_formatter(NullFormatter())
    axs[2].set_xlabel("$\\mu_{R}/\\mu_0$")

    # axs[3].plot(t, s3)
    axs[3].set_xscale("log")
    axs[3].set_xlim(np.min(mf), np.max(mf))
    axs[3].set_xticks([1.0])
    axs[3].xaxis.set_minor_formatter(NullFormatter())
    axs[3].set_xlabel("$\\mu_{F}/\\mu_0$")

    # axs[4].plot(t, s3)
    axs[4].set_xscale("log")
    axs[4].set_xlim(np.min(mf), np.max(mf))
    axs[4].set_xticks([1.0])
    axs[4].xaxis.set_minor_formatter(NullFormatter())
    axs[4].set_xlabel("$\\mu_{R}/\\mu_0$")

    axs[0].legend(handletextpad=-0.0, handlelength=0, fancybox=False)
    axs[1].legend(handletextpad=-0.0, handlelength=0, fancybox=False)
    axs[2].legend(handletextpad=-0.0, handlelength=0, fancybox=False)
    axs[3].legend(handletextpad=-0.0, handlelength=0, fancybox=False)
    axs[4].legend(handletextpad=-0.0, handlelength=0, fancybox=False)
    fig.legend()
    fig.legends = []
    fig.legend(handles=lines, labels=labels, loc="center right")
    plt.subplots_adjust(right=0.84)
    # plt.show()
    mpl.rcParams["axes.prop_cycle"] = cycle_safe


def central_scale_plot(
    dict_list, vl, cont=False, error=True, yscale=1.0, unit="pb", yaxis=None
):
    """Creates a scale variance plot with 3 panels (ystacked)."""
    global fig, axs
    if not cont:
        fig, axs = plt.subplots(3, 1, figsize=(12, 8), sharex=True)
        # Remove horizontal space between axes
        fig.subplots_adjust(hspace=0)

    mr = dict_list["mu_r"]
    mf = dict_list["mu_f"]

    for v in vl:
        mv = dict_list[v] * yscale

        mask = mf == mr
        l = err_plt(axs[0], mf[mask], mv[mask], label=v, error=error)

        mask = mf == 1.0
        l = err_plt(axs[1], mr[mask], mv[mask], error=error)

        mask = mr == 1.0
        l = err_plt(axs[2], mf[mask], mv[mask], error=error)

    if not cont:
        axs[0].plot([], [], " ", label="$\\mu_R=\\mu_F=\\mu$")
        axs[1].plot([], [], " ", label="$\\mu_F=\\mu_0$, $\\mu_R=\\mu$")
        axs[2].plot([], [], " ", label="$\\mu_R=\\mu_0$, $\\mu_F=\\mu$")

    axs[1].set_ylabel("$\\sigma$ [" + unit + "]" if yaxis is None else yaxis)

    axs[0].set_xscale("log")
    # axs[0].set_xlim(np.min(mf), np.max(mf))
    axs[2].set_xlabel("$\\mu/\\mu_0$")

    axs[0].legend()
    axs[1].legend()
    axs[2].legend()
    # plt.show()


def init_double_plot(
    figsize=(6, 8), sharex=True, sharey=False, gridspec_kw={"height_ratios": [3, 1]}
):
    """Initialze subplot for Ratio/K plots with another figure below."""
    fig, axs = plt.subplots(
        2, 1, figsize=figsize, sharex=sharex, sharey=sharey, gridspec_kw=gridspec_kw
    )
    # Remove horizontal space between axes
    fig.subplots_adjust(hspace=0)
    return fig, axs



import hepi

print(hepi.__version__)
import numpy as np

from hepi.input import set_input_dir, set_output_dir
from hepi.run import nnllfast as nnll

# set the folder where the resummino binary can be found either in either ./{,bin,bin/build}/resummino
nnll.set_path("nnll-fast-1.1")
set_input_dir("./tests/input/")
set_output_dir("./tests/output/")


def test_nnllfast():
    params = [
        "mastercode_with_gm2.in",  # baseline slha file in the relative ./output folder by default unless set_output_dir was used
    ]
    pss = [
        (1000001, 1000021),  # Final state particles for resummino to run
    ]

    for pa, pb in pss:
        for param in params:
            # All the inputs Order, CMS in GeV, particle 1, particle 2, slha, pdf_lo, pdf_nlo,mu_f, mu_r
            i = hepi.Input(
                hepi.Order.aNNLO_PLUS_NNLL,
                13000,
                pa,
                pb,
                param,
                "PDF4LHC15",
                "PDF4LHC15",
                1.0,
                1.0,
                id=1,
            )
            li = [i]  # li is our list of inputs that we want resummino to run
            li = hepi.mass_scan(li, pb, np.linspace(2000, 2000, 1))
            li = hepi.mass_scan(
                li, pa, np.linspace(1000, 2000, 16)
            )  # we scan the slepton mass from 100 to 1000 at 15 equidistant points
            nnll.run(li, skip=False, n_jobs=1)



import hepi

print(hepi.__version__)
import numpy as np

from hepi.input import set_input_dir, set_output_dir

set_input_dir("./tests/input/")
set_output_dir("./tests/output/")
from hepi.run import softsusy as ss


def test_softsusy():
    for sq in [2000002, 1000002]:
        for pdf, nlopdf in [("CT14lo", "CT14lo")]:
            li = [
                hepi.Input(
                    hepi.Order.LO,
                    13000,
                    sq,
                    1000022,
                    "pMSSM11_best_fit_LHC13-SUSYHIT.txt",
                    pdf,
                    nlopdf,
                    1.0,
                    1.0,
                    id="test",
                )
            ]
            li = hepi.slha_scan_rel(
                li,
                lambda r: [["EXTPAR", 1, 510], ["EXTPAR", 2, r]],
                np.linspace(470.0, 530.0, 32),
            )
            ss.run(li)



import pandas as pd
from smpl import interpolate as ip


# TODO fix dependent(mu) for new masses
def interpolate_1d(df, x, y, xrange, only_interpolation=True,**kwargs):
    """
    Last key is the value to be interpolated, while the rest are cooridnates.

    Args:
        df (pandas.DataFrame): results
    """
    f = ip.interpolate(df[x], df[y],**kwargs)
    a = []
    for xr in xrange:
        c = df.head(1)
        c[x] = xr
        c[y] = f(xr)
        a += [c]
    if only_interpolation:
        return pd.concat(a)
    else:
        return pd.concat([df, *a])


def interpolate_2d(df, x, y, z, xrange, yrange, only_interpolation=True, **kwargs):
    """
    Last key is the value to be interpolated, while the rest are cooridnates.

    Args:
        df (pandas.DataFrame): results
    """
    f = ip.interpolate(df[x], df[y], df[z], **kwargs)
    a = []
    for i in range(len(xrange)):
        xr = xrange[i]
        yr = yrange[i]
        zr = f(xr, yr)
        c = df.head(1).copy()
        c[x] = xr
        c[y] = yr
        c[z] = zr
        a += [c]
    if only_interpolation:
        return pd.concat(a)
    else:
        return pd.concat([df, *a])



from enum import IntEnum


class Order(IntEnum):
    """
    Computation orders.
    """

    LO = 0
    """Leading Order"""
    NLO = 1
    """Next-to-Leading Order"""
    NLO_PLUS_NLL = 2
    """Next-to-Leading Order plus Next-to-Leading Logarithms"""
    aNNLO_PLUS_NNLL = 3
    """Approximate Next-to-next-to-Leading Order plus Next-to-next-to-Leading Logarithms"""


def replace_macros(s: str) -> str:
    return s.replace("_PLUS_", "+").replace(" ", "\\ ")


def xsec_to_order(s: str):
    if s == "NNLOapprox+NNLL":
        return Order.aNNLO_PLUS_NNLL
    elif s == "NLO+NLL":
        return Order.NLO_PLUS_NLL
    elif s == "NLO":
        return Order.NLO
    elif s == "LO":
        return Order.LO
    else:
        raise ValueError("Unknown Order '" + s + "', not supported by HEPi.")


def order_to_string(o: Order, json_style=False, no_macros=False) -> str:
    ret = ""
    if o == Order.LO:
        ret = "LO"
    elif o == Order.NLO:
        ret = "NLO"
    elif o == Order.NLO_PLUS_NLL:
        ret = "NLO_PLUS_NLL"
    elif o == Order.aNNLO_PLUS_NNLL:
        if json_style:
            ret = "NNLOapprox+NNLL"
        else:
            ret = "aNNLO_PLUS_NNLL"
    else:
        raise ValueError("Order '" + o + "' not supported by HEPi.")
    if no_macros:
        return replace_macros(ret)
    return ret



#!/usr/bin/env python
# Interpolate data (loaded from json)
import argparse
import sys
import urllib.request
import warnings
from os import path

import numpy as np
import uncertainties.unumpy as unp
import validators
from smpl import interpolate as ip

from hepi.input import order_to_string
from hepi.load import load_json_with_metadata
from hepi.order import replace_macros

# from hepi import data

try:
    import readline
except ImportError:
    pass

import hepi.data

unv = unp.nominal_values
usd = unp.std_devs


def main():
    parser = argparse.ArgumentParser(
        description="""Interpolate data (loaded from json) in the format:
ID | Central value | error up | error down | error scale up | error scale down | error pdf up | error pdf down """
    )
    parser.add_argument(
        "json", type=str, nargs="*", help="url/file/string/name", default=[]
    )
    parser.add_argument(
        "-l", "--list", action="store_true", help="list installed grids", default=False
    )
    parser.add_argument(
        "-p", "--plot", action="store_true", help="plot listed files", default=False
    )
    parser.add_argument(
        "-i",
        "--info",
        action="store_true",
        help="show info of listed files",
        default=False,
    )
    parser.add_argument(
        "-s",
        "--size",
        type=int,
        help="number of points after interpolation",
        default=50,
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="output file (default: None)",
        default=None,
    )
    parser.add_argument(
        "-d",
        "--data",
        type=str,
        help="just prints the data in given format: json, twiki, latex",
        default=None,
    )
    args = parser.parse_args()

    fs = []

    if args.list:
        print("Available grids:")
        for f in hepi.data.list_files():
            print(f)
        return
    if args.plot:
        # on the fly import to speed up hepi-fast
        from matplotlib import pyplot as plt
        from smpl import data

        from hepi.interpolate import interpolate_2d
        from hepi.plot import combined_plot, mapplot

    for xs, j in enumerate(args.json):
        if path.exists(j):
            with open(j) as f:
                df, d, _ = load_json_with_metadata(f)
        elif j in hepi.data.list_files() or j + ".json" in hepi.data.list_files():
            with open(hepi.data.get_file(j.replace(".json", "") + ".json")) as f:
                df, d, _ = load_json_with_metadata(f)
        elif validators.url(j):
            with urllib.request.urlopen(j) as f:
                df, d, _ = load_json_with_metadata(f)
        else:
            warnings.warn("Unknown input: " + j)
            continue

        if args.info:
            print(
                "Process "
                + str(df["slha"].iloc[0])
                + " using "
                + str(df["pdf_nlo"].iloc[0])
                + " at energy "
                + str(df["energy"].iloc[0])
                + " GeV computed with "
                + str(df["runner"].iloc[0])
            )
            continue

        so = order_to_string(df["order"].iloc[0])

        if args.data == "json":
            hepi.write_json(df, df["order"].iloc[0], [d[0][0]], sys.stdout)
            continue

        if args.data == "twiki":
            hepi.write_twiki_header(d[0][0], sys.stdout, unit="fb")
            hepi.write_twiki(df, d[0][0], so, sys.stdout, yscale=1000)
            continue

        if args.data == "latex":
            cen = df[so+ "_NOERR"]
            df[so] = df[so+ "_NOERR"]

            df[so + "_SCALE_ERRPLUS"] = (unv(df[so+ "_SCALE"])-cen + usd(df[so+ "_SCALE"]))
            df[so + "_SCALE_ERRMINUS"] = (unv(df[so+ "_SCALE"])-cen - usd(df[so+ "_SCALE"]))
            df[so + "_PDF_ERRPLUS"] = (unv(df[so+ "_PDF"])-cen + usd(df[so+ "_PDF"]))
            df[so + "_PDF_ERRMINUS"] = (unv(df[so+ "_PDF"])-cen - usd(df[so+ "_PDF"]))

            hepi.write_latex_table_transposed_header(df, so,sys.stdout, d[0][0])
            hepi.write_latex_table_transposed(df, so,sys.stdout, d[0][0],yscale=1000)
            continue

        if len(d) == 1:
            dat = [df[d[0][0]]]
            interpolator = "cubic"
            if args.plot:
                combined_plot(
                    df,
                    d[0][0],
                    so,
                    tight=False,
                    plot_data=False,
                    logy=True,
                    interpolate=True,
                    xaxis=d[0][0],
                    interpolator="cubic",
                    pre=np.log,
                    post=np.exp,
                    init=False,
                    label=df["slha"].iloc[0]
                    + " @ "
                    + replace_macros(order_to_string(df["order"].iloc[0]))
                    + " with "
                    + str(df["pdf_nlo"].iloc[0])
                    + " at $\\sqrt{S} = "
                    + str(int(df["energy"].iloc[0]) / 1000)
                    + "$ TeV",
                )
                # hepi.title(li,axe=axs[xs,0],cms_energy=True,pdf_info=False)
        elif len(d) == 2:
            if xs == 0 and args.plot:
                fig, axs = plt.subplots(
                    len(args.json), 2
                )  # Remove horizontal space between axes
                if len(args.json) == 1:
                    axs = np.array([axs])

            dat = [df[d[0][0]], df[d[1][0]]]
            interpolator = "linearnd"
            if args.plot:
                x = np.around(np.linspace(dat[0].min(), dat[0].max(), args.size), 0)
                y = np.around(np.linspace(dat[1].min(), dat[1].max(), args.size), 0)
                xx, yy = data.flatmesh(x, y)
                dll = interpolate_2d(
                    df,
                    d[0][0],
                    d[1][0],
                    so + "_COMBINED",
                    xx,
                    yy,
                    interpolator="linearnd",
                    interpolate_lower_uncertainty=False,
                    pre=np.log,
                    post=np.exp,
                )
                mapplot(
                    dll,
                    d[0][0],
                    d[1][0],
                    so + "_COMBINED",
                    axes=axs[xs, 0],
                    xaxis=d[0][0],
                    yaxis=d[1][0],
                    zaxis="$\\sigma$ [pb]",
                    fill_missing=False,
                    init=True,
                )
                dlln = interpolate_2d(
                    df,
                    d[0][0],
                    d[1][0],
                    so + "_NOERR",
                    xx,
                    yy,
                    interpolator="linearnd",
                    interpolate_lower_uncertainty=False,
                    pre=np.log,
                    post=np.exp,
                )
                dll[so + "_NOERR"] = dlln[so + "_NOERR"]
                for m_x in y[::5]:
                    mask = (
                        (dll[d[1][0]] == m_x)
                        & (dll[so + "_NOERR"].notnull())
                        & (dll[so + "_COMBINED"].notnull())
                        & (dll[so + "_NOERR"] > 0)
                        & (dll[so + "_COMBINED"] > 0)
                    )
                    hepi.combined_plot(
                        dll[mask],
                        d[0][0],
                        so,
                        alpha=0.1,
                        axes=axs[xs, 1],
                        tight=False,
                        init=False,
                        interpolator="cubic",
                        xaxis=d[0][0],
                        label=str(m_x),
                        cont=True,
                        plot_data=False,
                        pre=np.log,
                        post=np.exp,
                        interpolate=True,
                    )
                    plt.legend(title=d[1][0])

        else:
            raise ValueError("Only 1 or 2 dimensions supported.")
        if not args.plot:
            f_noerr = ip.interpolate(
                *dat,
                df[so + "_NOERR"],
                interpolator=interpolator,
                pre=np.log,
                post=np.exp,
                interpolate_lower_uncertainty=False,
            )
            f_combined = ip.interpolate(
                *dat,
                df[so + "_COMBINED"],
                interpolator=interpolator,
                pre=np.log,
                post=np.exp,
                interpolate_lower_uncertainty=False,
            )
            if so + "_PDF" in df.columns:
                f_pdf = ip.interpolate(
                    *dat,
                    df[so + "_PDF"],
                    interpolator=interpolator,
                    pre=np.log,
                    post=np.exp,
                    interpolate_lower_uncertainty=False,
                )
            else:
                f_pdf = f_noerr
            if so + "_SCALE" in df.columns:
                f_scale = ip.interpolate(
                    *dat,
                    df[so + "_SCALE"],
                    interpolator=interpolator,
                    pre=np.log,
                    post=np.exp,
                    interpolate_lower_uncertainty=False,
                )
            else:
                f_scale = f_noerr

            fs = [*fs, (f_noerr, f_combined, f_pdf, f_scale)]
    if args.plot:
        from matplotlib import pyplot as plt

        if args.output is not None:
            plt.savefig(args.output)
        plt.show()
    else:
        if len(fs) != 0:
            try:
                for line in sys.stdin:
                    for i, (f_noerr, f_combined, f_pdf, f_scale) in enumerate(fs):
                        arg = [float(a) for a in line.split(" ")]
                        vn = f_noerr(*arg)
                        vs = f_scale(*arg)
                        vp = f_pdf(*arg)
                        vc = f_combined(*arg)
                        print(
                            i,
                            *arg,
                            vn,
                            -vn + unv(vc) + usd(vc),
                            -vn + unv(vc) - usd(vc),
                            -vn + unv(vs) + usd(vs),
                            -vn + unv(vs) - usd(vs),
                            -vn + unv(vp) + usd(vp),
                            -vn + unv(vp) - usd(vp),
                        )
            except KeyboardInterrupt:
                # stops the inifinite input loop
                pass
